"use strict";
/*!
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const uuid_1 = require("uuid");
const constants_1 = require("../constants");
const error_1 = __importStar(require("../error"));
const interceptor_1 = __importStar(require("../interceptor"));
const query_1 = require("../query");
const sprocs = __importStar(require("../sproc"));
const task_1 = require("../task");
const batchIterator_1 = __importDefault(require("../utils/batchIterator"));
const computeNextRun_1 = __importDefault(require("../utils/computeNextRun"));
const isValidNextRun_1 = __importDefault(require("../utils/isValidNextRun"));
const cosmos_1 = __importDefault(require("./cosmos"));
const listener_1 = __importDefault(require("./listener"));
const scoped_1 = __importDefault(require("./scoped"));
// Async iterator polyfill for Node <10
if (!Symbol.asyncIterator) {
    Symbol.asyncIterator = Symbol.for('Symbol.asyncIterator');
}
/**
 * The primary client for creating, updating, querying and processing tasks. It
 * wraps an Azure Cosmos DB collection, which is used to store the tasks. It's
 * recommended that you create one using the TaskClient.create method instead of
 * using the constructor. This will set up the database, collection and
 * necessary stored procedures for you.
 *
 * @public
 */
class TaskClient {
    /**
     * Create a client using a pre-initialized Azure Cosmos DB client. Usage of
     * this API is only recommended for advanced users. Most users should use
     * the TaskClient.create method instead.
     *
     * @param cosmosClient    - Client created from a {@link
     *                          @azure/cosmos#Container} that is scoped to the
     *                          collection/container that the tasks client
     *                          should use.
     * @param options         - Client creation options
     *
     * @internal
     */
    constructor(cosmosClient, options) {
        this._client = cosmosClient;
        this._options = Object.assign({ lockDurationMs: 30000, maxExecutionTimeMs: 30 * 60 * 1000, pollIntervalMs: 5000, maxUpdateParallelism: 10, interceptors: {}, retries: constants_1.INTERNAL_RETRY_OPTIONS }, options);
        this._interceptor = new interceptor_1.InterceptorProcessor(this._options.interceptors);
    }
    /**
     * Initializes a task client using the provided account information,
     * creating the specified database and collection if necessary.
     *
     * @param account     - Azure Cosmos DB account url
     * @param database    - Azure Cosmos DB database name
     * @param collection  - Azure Cosmos DB collection/container name
     * @param key         - Azure Cosmos DB account key
     * @param options     - Client creation options
     *
     * @returns Promise containing the initialized client
     *
     * @public
     */
    static async create(account, database, collection, key, options) {
        const cosmosClient = await cosmos_1.default.create(account, database, collection, key, {
            partitionKey: {
                paths: ['/config/type']
            },
            defaultTtl: -1
        });
        const taskClient = new TaskClient(cosmosClient, options);
        await taskClient.registerSprocs();
        return taskClient;
    }
    /**
     * Returns a client with all operations automatically scoped to the provided
     * task type. This is useful if you have an area where you need to perform
     * many related operations for a single task type and don't want to have to
     * specify it each time.
     *
     * @param type    - Task type to apply to all operations within the scoped
     *                  client.
     *
     * @public
     */
    type(type) {
        return new scoped_1.default(this, type);
    }
    /**
     * Creates a task and saves it to the database.
     *
     * @param type    - Task type
     * @param payload - Payload containing the data for the task
     * @param options - Options to configure the newly created task
     *
     * @public
     */
    async create(type, payload, options) {
        const id = uuid_1.v4();
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.Create, this._client.documentRef(type, id), type, async () => {
            const definedOptions = options || {};
            const now = Date.now();
            // Calculate nextRunTime
            let nextRunTime = computeNextRun_1.default(definedOptions.interval, undefined, definedOptions.lastRunTime);
            if (definedOptions.scheduledTime) {
                nextRunTime = isValidNextRun_1.default(definedOptions.scheduledTime.getTime(), definedOptions.lastRunTime);
            }
            const task = {
                id,
                config: {
                    type,
                    enabled: definedOptions.enabled !== undefined
                        ? definedOptions.enabled
                        : true,
                    createTime: now,
                    lockedUntilTime: 0,
                    nextRunTime,
                    deliveries: 0,
                    attempts: 0,
                    runs: 0,
                    interval: definedOptions.interval,
                    lastRunTime: definedOptions.lastRunTime !== undefined
                        ? definedOptions.lastRunTime.getTime()
                        : undefined,
                    ttlMs: definedOptions.ttlMs,
                    maxExecutionTimeMs: definedOptions.maxExecutionTimeMs
                },
                payload
            };
            const response = await this._client.createItem(task);
            return Object.assign(Object.assign({}, response), { result: task_1.TaskImpl.create(this._client, this._interceptor, response.result) });
        });
    }
    /**
     * Retrieves the task of the given type with the provided id, if one exists.
     * If the task is not found, undefined is returned.
     *
     * @param type    - Task type
     * @param id      - The task's id
     *
     * @public
     */
    async get(type, id) {
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.Get, this._client.documentRef(type, id), type, async () => {
            const response = await this._client.getItem(id, type);
            return Object.assign(Object.assign({}, response), { result: response.result
                    ? task_1.TaskImpl.create(this._client, this._interceptor, response.result)
                    : undefined });
        });
    }
    /**
     * Retrieves all tasks of the given type, paged using the provided list
     * options.
     *
     * @param type    - Task type
     * @param options - Options controlling which tasks to retrieve
     *
     * @public
     */
    async list(type, options = {}) {
        const skip = options.skip || 0;
        const top = options.top || constants_1.DEFAULT_PAGE_SIZE;
        const query = query_1.buildQuery({
            offset: skip,
            limit: top,
            filter: task_1.typeFilter(type, options.filter),
            sort: options.sortExpression,
            sortOrder: options.sortOrder
        });
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.List, this._client.containerRef, type, async () => {
            const response = await this._client.queryItemsArray(query);
            return Object.assign(Object.assign({}, response), { result: response.result.map(doc => task_1.TaskImpl.create(this._client, this._interceptor, doc)) });
        });
    }
    /**
     * Retrieves a single page from the database for tasks of the given type,
     * optionally starting from the provided continuation token.
     *
     * @param type    - Task type
     * @param options - Options controlling which tasks to retrieve
     *
     * @public
     */
    async listPaged(type, options = {}) {
        const query = query_1.buildQuery({
            filter: task_1.typeFilter(type, options.filter),
            sort: options.sortExpression,
            sortOrder: options.sortOrder
        });
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.ListPaged, this._client.containerRef, type, async () => {
            const response = await this._client.queryItemsPage(query, {
                continuation: options.continuation,
                pageSize: options.pageSize || constants_1.DEFAULT_PAGE_SIZE
            });
            return Object.assign(Object.assign({}, response), { result: this._addContinuation(response.result.map(doc => task_1.TaskImpl.create(this._client, this._interceptor, doc)), response.continuation) });
        });
    }
    /**
     * Retrieves all tasks across all types of tasks, paged using the provided
     * list options.
     *
     * @remarks
     *
     * Performing operations across multiple types results in cross partition
     * queries to Cosmos DB. This can cause a significant performance impact,
     * especially when paging and/or sorting results. Use this API with care.
     *
     * @param options - Options controlling which tasks to retrieve
     *
     * @public
     */
    async listAll(options = {}) {
        const skip = options.skip || 0;
        const top = options.top || constants_1.DEFAULT_PAGE_SIZE;
        const query = query_1.buildQuery({
            offset: skip,
            limit: top,
            filter: options.filter,
            sort: options.sortExpression,
            sortOrder: options.sortOrder
        });
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.ListAll, this._client.containerRef, undefined, async () => {
            const response = await this._client.queryItemsArray(query);
            return Object.assign(Object.assign({}, response), { result: response.result.map(doc => task_1.TaskImpl.create(this._client, this._interceptor, doc)) });
        });
    }
    /**
     * Retrieves all tasks of the given type with the entire payload omitted by
     * default. This is primarily useful if you have tasks with a large amount
     * of data in the payload that you don't need to see in the listed results
     * and you want to save cost/memory.
     *
     * @remarks
     *
     * Results are paged, filtered and sorted using the provided options. You
     * may also specify certain properties of the payload you want to see in the
     * returned task through the options.
     *
     * @param type    - Task type
     * @param options - Options controlling which tasks to retrieve
     *
     * @public
     */
    async listSummary(type, options = {}) {
        const skip = options.skip || 0;
        const top = options.top || constants_1.DEFAULT_PAGE_SIZE;
        const query = query_1.buildQuery({
            offset: skip,
            limit: top,
            projection: task_1.summaryProjection(options.project),
            filter: task_1.typeFilter(type, options.filter),
            sort: options.sortExpression,
            sortOrder: options.sortOrder
        });
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.ListSummary, this._client.containerRef, type, async () => {
            const response = await this._client.queryItemsArray(query);
            return Object.assign(Object.assign({}, response), { result: response.result.map(doc => task_1.ReadonlyTaskImpl.create(this._client, this._interceptor, doc)) });
        });
    }
    /**
     * Retrieves a single page of tasks of the given type with the entire payload
     * omitted by default, optionally starting from the provided continuation
     * token. This is primarily useful if you have tasks with a large amount of
     * data in the payload that you don't need to see in the listed results and
     * you want to save cost/memory.
     *
     * @remarks
     *
     * Results are paged, filtered and sorted using the provided options. You
     * may also specify certain properties of the payload you want to see in the
     * returned task through the options.
     *
     * @param type    - Task type
     * @param options - Options controlling which tasks to retrieve
     *
     * @public
     */
    async listSummaryPaged(type, options = {}) {
        const query = query_1.buildQuery({
            projection: task_1.summaryProjection(options.project),
            filter: task_1.typeFilter(type, options.filter),
            sort: options.sortExpression,
            sortOrder: options.sortOrder
        });
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.ListSummaryPaged, this._client.containerRef, type, async () => {
            const response = await this._client.queryItemsPage(query, {
                continuation: options.continuation,
                pageSize: options.pageSize || constants_1.DEFAULT_PAGE_SIZE
            });
            return Object.assign(Object.assign({}, response), { result: this._addContinuation(response.result.map(doc => task_1.ReadonlyTaskImpl.create(this._client, this._interceptor, doc)), response.continuation) });
        });
    }
    /**
     * Retrieves all tasks across all types with the entire payload omitted by
     * default. This is primarily useful if you have tasks with a large amount
     * of data in the payload that you don't need to see in the listed results
     * and you want to save cost/memory.
     *
     * @remarks
     *
     * Results are paged, filtered and sorted using the provided options. You
     * may also specify certain properties of the payload you want to see in the
     * returned task through the options.
     *
     * Performing operations across multiple types results in cross partition
     * queries to Cosmos DB. This can cause a significant performance impact,
     * especially when paging and/or sorting results. Use this API with care.
     *
     * @param options - Options controlling which tasks to retrieve
     *
     * @public
     */
    async listAllSummary(options = {}) {
        const skip = options.skip || 0;
        const top = options.top || constants_1.DEFAULT_PAGE_SIZE;
        const query = query_1.buildQuery({
            offset: skip,
            limit: top,
            projection: task_1.summaryProjection(options.project),
            filter: options.filter,
            sort: options.sortExpression,
            sortOrder: options.sortOrder
        });
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.ListAllSummary, this._client.containerRef, undefined, async () => {
            const response = await this._client.queryItemsArray(query);
            return Object.assign(Object.assign({}, response), { result: response.result.map(doc => task_1.ReadonlyTaskImpl.create(this._client, this._interceptor, doc)) });
        });
    }
    /**
     * Returns an async iterator for iterating over all tasks of the given type,
     * optionally filtered/sorted using the provided options object.
     *
     * @remarks
     *
     * _NOTE: Asynchronous iterators are a new feature of Javascript and are
     * only available without a polyfill starting in Node.js 10. You can make
     * use of async iterators in prior versions of Node.js by using TypeScript
     * or Babel to downcompile the code for you:
     * https://node.green/#ES2018-features-Asynchronous-Iterators_
     *
     * @param type    - Task type
     * @param options - Options controlling which tasks to retrieve and the
     *                  order in which to receive them
     *
     * @public
     */
    iterate(type, options = {}) {
        return __asyncGenerator(this, arguments, function* iterate_1() {
            var e_1, _a;
            const query = query_1.buildQuery({
                filter: task_1.typeFilter(type, options.filter),
                sort: options.sortExpression,
                sortOrder: options.sortOrder
            });
            const iterator = this._client.queryItemsIterator(query);
            const mappedIterator = this._iterate(interceptor_1.default.TaskClientOperation.Iterate, type, iterator, doc => task_1.TaskImpl.create(this._client, this._interceptor, doc));
            try {
                for (var mappedIterator_1 = __asyncValues(mappedIterator), mappedIterator_1_1; mappedIterator_1_1 = yield __await(mappedIterator_1.next()), !mappedIterator_1_1.done;) {
                    const task = mappedIterator_1_1.value;
                    yield yield __await(task);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (mappedIterator_1_1 && !mappedIterator_1_1.done && (_a = mappedIterator_1.return)) yield __await(_a.call(mappedIterator_1));
                }
                finally { if (e_1) throw e_1.error; }
            }
        });
    }
    /**
     * Returns an async iterator for iterating over all tasks across all types,
     * optionally filtered/sorted using the provided options object.
     *
     * @remarks
     *
     * Performing operations across multiple types results in cross partition
     * queries to Cosmos DB. This can cause a significant performance impact,
     * especially when paging and/or sorting results. Use this API with care.
     *
     * _NOTE: Asynchronous iterators are a new feature of Javascript and are
     * only available without a polyfill starting in Node.js 10. You can make
     * use of async iterators in prior versions of Node.js by using TypeScript
     * or Babel to downcompile the code for you:
     * https://node.green/#ES2018-features-Asynchronous-Iterators_
     *
     * @param options - Options controlling which tasks to retrieve and the
     *                  order in which to receive them
     *
     * @public
     */
    iterateAll(options = {}) {
        return __asyncGenerator(this, arguments, function* iterateAll_1() {
            var e_2, _a;
            const query = query_1.buildQuery({
                filter: options.filter,
                sort: options.sortExpression,
                sortOrder: options.sortOrder
            });
            const iterator = this._client.queryItemsIterator(query);
            const mappedIterator = this._iterate(interceptor_1.default.TaskClientOperation.IterateAll, undefined, iterator, doc => task_1.TaskImpl.create(this._client, this._interceptor, doc));
            try {
                for (var mappedIterator_2 = __asyncValues(mappedIterator), mappedIterator_2_1; mappedIterator_2_1 = yield __await(mappedIterator_2.next()), !mappedIterator_2_1.done;) {
                    const task = mappedIterator_2_1.value;
                    yield yield __await(task);
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (mappedIterator_2_1 && !mappedIterator_2_1.done && (_a = mappedIterator_2.return)) yield __await(_a.call(mappedIterator_2));
                }
                finally { if (e_2) throw e_2.error; }
            }
        });
    }
    /**
     * Returns an async iterator for iterating over all tasks of the given type
     * with the entire payload omitted by default. This is primarily useful if
     * you have tasks with a large amount of data in the payload that you don't
     * need to see in the listed results and you want to save cost/memory.
     *
     * @remarks
     *
     * Results are filtered and sorted using the provided options. You may also
     * specify certain properties of the payload you want to see in the returned
     * task through the options.
     *
     * _NOTE: Asynchronous iterators are a new feature of Javascript and are
     * only available without a polyfill starting in Node.js 10. You can make
     * use of async iterators in prior versions of Node.js by using TypeScript
     * or Babel to downcompile the code for you:
     * https://node.green/#ES2018-features-Asynchronous-Iterators_
     *
     * @param type    - Task type
     * @param options - Options controlling which tasks to retrieve and the
     *                  order in which to receive them
     * @public
     */
    iterateSummary(type, options = {}) {
        return __asyncGenerator(this, arguments, function* iterateSummary_1() {
            var e_3, _a;
            const query = query_1.buildQuery({
                projection: task_1.summaryProjection(options.project),
                filter: task_1.typeFilter(type, options.filter),
                sort: options.sortExpression,
                sortOrder: options.sortOrder
            });
            const iterator = this._client.queryItemsIterator(query);
            const mappedIterator = this._iterate(interceptor_1.default.TaskClientOperation.IterateSummary, type, iterator, doc => task_1.ReadonlyTaskImpl.create(this._client, this._interceptor, doc));
            try {
                for (var mappedIterator_3 = __asyncValues(mappedIterator), mappedIterator_3_1; mappedIterator_3_1 = yield __await(mappedIterator_3.next()), !mappedIterator_3_1.done;) {
                    const task = mappedIterator_3_1.value;
                    yield yield __await(task);
                }
            }
            catch (e_3_1) { e_3 = { error: e_3_1 }; }
            finally {
                try {
                    if (mappedIterator_3_1 && !mappedIterator_3_1.done && (_a = mappedIterator_3.return)) yield __await(_a.call(mappedIterator_3));
                }
                finally { if (e_3) throw e_3.error; }
            }
        });
    }
    /**
     * Returns an async iterator for iterating over all tasks across all types
     * with the entire payload omitted by default. This is primarily useful if
     * you have tasks with a large amount of data in the payload that you don't
     * need to see in the listed results and you want to save cost/memory.
     *
     * @remarks
     *
     * Results are filtered and sorted using the provided options. You may also
     * specify certain properties of the payload you want to see in the returned
     * task through the options.
     *
     * Performing operations across multiple types results in cross partition
     * queries to Cosmos DB. This can cause a significant performance impact,
     * especially when paging and/or sorting results. Use this API with care.
     *
     * _NOTE: Asynchronous iterators are a new feature of Javascript and are
     * only available without a polyfill starting in Node.js 10. You can make
     * use of async iterators in prior versions of Node.js by using TypeScript
     * or Babel to downcompile the code for you:
     * https://node.green/#ES2018-features-Asynchronous-Iterators_
     *
     * @param type    - Task type
     * @param options - Options controlling which tasks to retrieve and the
     *                  order in which to receive them
     * @public
     */
    iterateAllSummary(options = {}) {
        return __asyncGenerator(this, arguments, function* iterateAllSummary_1() {
            var e_4, _a;
            const query = query_1.buildQuery({
                projection: task_1.summaryProjection(options.project),
                filter: options.filter,
                sort: options.sortExpression,
                sortOrder: options.sortOrder
            });
            const iterator = this._client.queryItemsIterator(query);
            const mappedIterator = this._iterate(interceptor_1.default.TaskClientOperation.IterateAllSummary, undefined, iterator, doc => task_1.ReadonlyTaskImpl.create(this._client, this._interceptor, doc));
            try {
                for (var mappedIterator_4 = __asyncValues(mappedIterator), mappedIterator_4_1; mappedIterator_4_1 = yield __await(mappedIterator_4.next()), !mappedIterator_4_1.done;) {
                    const task = mappedIterator_4_1.value;
                    yield yield __await(task);
                }
            }
            catch (e_4_1) { e_4 = { error: e_4_1 }; }
            finally {
                try {
                    if (mappedIterator_4_1 && !mappedIterator_4_1.done && (_a = mappedIterator_4.return)) yield __await(_a.call(mappedIterator_4));
                }
                finally { if (e_4) throw e_4.error; }
            }
        });
    }
    /**
     * Compute the number of tasks of the given type that currently exist,
     * optionally filtered to a custom group of tasks within the type.
     *
     * @param type    - Task type
     * @param filter  - Query filter specifying which tasks within the provided
     *                  type to include
     */
    async count(type, filter) {
        const query = query_1.buildQuery({
            filter: task_1.typeFilter(type, filter),
            rawProjection: 'VALUE COUNT(1)'
        });
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.Count, this._client.containerRef, type, async () => {
            const response = await this._client.queryItemsArray(query);
            return Object.assign(Object.assign({}, response), { result: response.result[0] });
        });
    }
    /**
     * Compute the number of tasks across all types that currently exist,
     * optionally filtered to a custom group of tasks.
     *
     * @remarks
     *
     * Performing operations across multiple types results in cross partition
     * queries to Cosmos DB. This can cause a significant performance impact,
     * especially when paging and/or sorting results. Use this API with care.
     *
     * @param filter  - Query filter specifying which tasks to include
     */
    async countAll(filter) {
        const query = query_1.buildQuery({
            filter,
            rawProjection: 'VALUE COUNT(1)'
        });
        return this._interceptor.client(this, interceptor_1.default.TaskClientOperation.Count, this._client.containerRef, undefined, async () => {
            const response = await this._client.queryItemsArray(query);
            return Object.assign(Object.assign({}, response), { result: response.result[0] });
        });
    }
    /**
     * Disables tasks of the provided type, optionally filtered to a custom group
     * of tasks within the type.
     *
     * @param type    - Task type
     * @param filter  - Query filter specifying which tasks within the provided
     *                  type to disable
     *
     * @public
     */
    async disable(type, filter) {
        await this._bulkUpdate(interceptor_1.default.TaskClientOperation.Disable, interceptor_1.default.TaskOperation.Disable, type, query_1.buildQuery({ filter: task_1.typeFilter(type, filter) }), async (task) => task.disable());
    }
    /**
     * Disables tasks across all types, optionally filtered to a custom group of
     * tasks.
     *
     * @remarks
     *
     * Performing operations across multiple types results in cross partition
     * queries to Cosmos DB. This can cause a significant performance impact,
     * especially when paging and/or sorting results. Use this API with care.
     *
     * @param filter  - Query filter specifying which tasks to disable
     *
     * @public
     */
    async disableAll(filter) {
        await this._bulkUpdate(interceptor_1.default.TaskClientOperation.DisableAll, interceptor_1.default.TaskOperation.Disable, undefined, query_1.buildQuery({ filter }), async (task) => task.disable());
    }
    /**
     * Enables tasks of the provided type, optionally filtered to a custom group
     * of tasks within the type.
     *
     * @param type    - Task type
     * @param filter  - Query filter specifying which tasks within the provided
     *                  type to enable
     * @param options - Task enable options
     *
     * @public
     */
    async enable(type, filter, options) {
        await this._bulkUpdate(interceptor_1.default.TaskClientOperation.Enable, interceptor_1.default.TaskOperation.Enable, type, query_1.buildQuery({ filter: task_1.typeFilter(type, filter) }), async (task) => task.enable(options));
    }
    /**
     * Enables tasks across all types, optionally filtered to a custom group
     * of tasks.
     *
     * @remarks
     *
     * Performing operations across multiple types results in cross partition
     * queries to Cosmos DB. This can cause a significant performance impact,
     * especially when paging and/or sorting results. Use this API with care.
     *
     * @param filter  - Query filter specifying which tasks to enable
     * @param options - Task enable options
     *
     * @public
     */
    async enableAll(filter, options) {
        await this._bulkUpdate(interceptor_1.default.TaskClientOperation.EnableAll, interceptor_1.default.TaskOperation.Enable, undefined, query_1.buildQuery({ filter }), async (task) => task.enable(options));
    }
    /**
     * Delete the task from the database. The operation is idempotent and will
     * succeed even if the task has already been deleted.
     *
     * @param type    - Task type
     * @param id      - The task's id
     *
     * @public
     */
    async deleteOne(type, id) {
        await this._interceptor.client(this, interceptor_1.default.TaskClientOperation.DeleteOne, this._client.documentRef(type, id), type, async () => this._client.deleteItem(id, type));
    }
    /**
     * Deletes tasks of the provided type, optionally filtered to a custom group
     * of tasks within the type.
     *
     * @param type    - Task type
     * @param filter  - Query filter specifying which tasks within the provided
     *                  type to delete
     *
     * @public
     */
    async delete(type, filter) {
        await this._bulkUpdate(interceptor_1.default.TaskClientOperation.Delete, interceptor_1.default.TaskOperation.Delete, type, query_1.buildQuery({ filter: task_1.typeFilter(type, filter) }), async (task) => task.delete());
    }
    /**
     * Deletes tasks across all types, optionally filtered to a custom group
     * of tasks.
     *
     * @param filter  - Query filter specifying which tasks to delete
     *
     * @public
     */
    async deleteAll(filter) {
        await this._bulkUpdate(interceptor_1.default.TaskClientOperation.DeleteAll, interceptor_1.default.TaskOperation.Delete, undefined, query_1.buildQuery({ filter }), async (task) => task.delete());
    }
    /**
     * Start listening for tasks of the provided type.
     *
     * @param type    - Task type
     * @param handler - Function to call with each task received for processing.
     *                  May be synchronous or asynchronous
     * @param options - Options for configuring the listener
     *
     * @public
     */
    listen(type, handler, options) {
        const opts = Object.assign(Object.assign({ maxActiveTasks: 100, lockDurationMs: this._options.lockDurationMs, pollIntervalMs: this._options.pollIntervalMs, maxExecutionTimeMs: this._options.maxExecutionTimeMs, refreshIntervalMs: constants_1.REFRESH_INTERVAL }, options), { retries: Object.assign(Object.assign({}, constants_1.DEFAULT_RETRY_OPTIONS), (options || {}).retries) });
        return new listener_1.default(this, this._client, this._interceptor, type, handler, opts);
    }
    /**
     * Ensures that the proper stored procedures are registered with Cosmos DB
     * for the client to function. You should never have to call this directly.
     *
     * @public
     */
    async registerSprocs() {
        await this._interceptor.client(this, interceptor_1.default.TaskClientOperation.RegisterSprocs, this._client.containerRef, undefined, async () => {
            let ruConsumption = 0;
            await Array.from(Object.values(sprocs)).map(async ({ id, fn }) => {
                try {
                    const result = await this._client.replaceSproc(id, fn);
                    ruConsumption += result.ruConsumption || 0;
                }
                catch (err) {
                    // 404 means we need to create it
                    if (error_1.default.is(err, error_1.ErrorCode.DATABASE_RESOURCE_NOT_FOUND)) {
                        try {
                            const result = await this._client.createSproc(id, fn);
                            ruConsumption += result.ruConsumption || 0;
                        }
                        catch (err) {
                            // 409s are okay at this point. Just means we
                            // were in a race with something else
                            if (error_1.default.is(err, error_1.ErrorCode.DATABASE_RESOURCE_ALREADY_EXISTS)) {
                                return;
                            }
                            throw err;
                        }
                    }
                    else {
                        throw err;
                    }
                }
            });
            return {
                ruConsumption,
                result: undefined
            };
        });
    }
    async _bulkUpdate(operation, taskOperation, type, query, op) {
        // TODO: do we want to return the count out?
        await this._interceptor.client(this, operation, this._client.containerRef, type, async () => {
            var e_5, _a;
            let taskCount = 0;
            let ruConsumption = 0;
            // We rig a custom interceptor that we use to build up our own
            // macro data for reporting to the end user. They should not
            // have to know/care that we are creating task instances and
            // running the operation on them.
            const interceptor = new interceptor_1.InterceptorProcessor({
                async task(context, next) {
                    // Allow all operations to execute
                    await next();
                    // We only care about our own operation
                    if (context.operation === taskOperation) {
                        ruConsumption += context.ruConsumption || 0;
                    }
                }
            });
            const iterator = await this._client.queryItemsIterator(query);
            try {
                for (var _b = __asyncValues(batchIterator_1.default(iterator, this._options.maxUpdateParallelism)), _c; _c = await _b.next(), !_c.done;) {
                    const results = _c.value;
                    for (const result of results) {
                        taskCount += 1;
                        ruConsumption += result.ruConsumption || 0;
                    }
                    await Promise.all(results
                        .map(({ result }) => task_1.TaskImpl.create(this._client, interceptor, result))
                        .map(async (task) => op(task)));
                }
            }
            catch (e_5_1) { e_5 = { error: e_5_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) await _a.call(_b);
                }
                finally { if (e_5) throw e_5.error; }
            }
            return {
                ruConsumption,
                result: taskCount
            };
        });
    }
    _iterate(operation, type, iterator, map) {
        const rawIterator = iterator[Symbol.asyncIterator]();
        return {
            [Symbol.asyncIterator]: () => ({
                next: async () => {
                    let value;
                    let done = false;
                    await this._interceptor.client(this, operation, this._client.containerRef, type, async () => {
                        const result = await rawIterator.next();
                        done = result.done || false;
                        if (!result.value) {
                            return {
                                result: undefined,
                                ruConsumption: undefined
                            };
                        }
                        value = map(result.value.result);
                        return {
                            result: value,
                            ruConsumption: result.value.ruConsumption
                        };
                    });
                    return {
                        value: value,
                        done
                    };
                }
            })
        };
    }
    _addContinuation(list, continuation) {
        const result = list;
        result.continuation = continuation;
        return result;
    }
}
exports.default = TaskClient;
//# sourceMappingURL=client.js.map