"use strict";
/*!
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const symbols_1 = require("../symbols");
/**
 * Convert a query expression into a query fragment.
 *
 * @param expression    The expression to serialize
 * @param paramIndex    Next available index to use for parameters (default: 0)
 */
function serializeExpression(expression, paramIndex = 0, isProjection = false) {
    if (isCustom(expression)) {
        // This is one of our custom expressions. Process it
        switch (expression[0]) {
            case symbols_1.PropSymbol.ANY:
            case symbols_1.PropSymbol.NUMERIC:
            case symbols_1.PropSymbol.BOOLEAN:
            case symbols_1.PropSymbol.STRING:
            case symbols_1.PropSymbol.SPATIAL:
            case symbols_1.PropSymbol.OBJECT:
            case symbols_1.PropSymbol.ARRAY:
                return serializers.property(expression, paramIndex, isProjection);
            case symbols_1.OpSymbol.AND:
            case symbols_1.OpSymbol.OR:
                return serializers.binaryLogical(expression, paramIndex, isProjection);
            case symbols_1.OpSymbol.NOT:
                return serializers.unaryLogical(expression, paramIndex, isProjection);
            case symbols_1.OpSymbol.ADD:
            case symbols_1.OpSymbol.SUBTRACT:
            case symbols_1.OpSymbol.MULTIPLY:
            case symbols_1.OpSymbol.DIVIDE:
            case symbols_1.OpSymbol.MODULO:
            case symbols_1.OpSymbol.BITWISE_AND:
            case symbols_1.OpSymbol.BITWISE_OR:
            case symbols_1.OpSymbol.XOR:
            case symbols_1.OpSymbol.LEFT_SHIFT:
            case symbols_1.OpSymbol.RIGHT_SHIFT:
            case symbols_1.OpSymbol.ZF_RIGHT_SHIFT:
            case symbols_1.OpSymbol.EQUAL:
            case symbols_1.OpSymbol.NOT_EQUAL:
            case symbols_1.OpSymbol.GREATER_THAN:
            case symbols_1.OpSymbol.GREATER_THAN_OR_EQUAL:
            case symbols_1.OpSymbol.LESS_THAN:
            case symbols_1.OpSymbol.LESS_THAN_OR_EQUAL:
            case symbols_1.OpSymbol.COALESCE:
                return serializers.binaryOperator(expression, paramIndex, isProjection);
            case symbols_1.OpSymbol.CONCATENATE:
                return serializers.concat(expression, paramIndex, isProjection);
            case symbols_1.OpSymbol.TERNARY:
                return serializers.ternaryOperator(expression, paramIndex, isProjection);
            case symbols_1.OpSymbol.IN:
                return serializers.inOp(expression, paramIndex, isProjection);
            case symbols_1.FnSymbol.PI:
                return serializers.pi(expression, paramIndex, isProjection);
            case symbols_1.FnSymbol.ABS:
            case symbols_1.FnSymbol.ACOS:
            case symbols_1.FnSymbol.ASIN:
            case symbols_1.FnSymbol.ATAN:
            case symbols_1.FnSymbol.CEILING:
            case symbols_1.FnSymbol.COS:
            case symbols_1.FnSymbol.COT:
            case symbols_1.FnSymbol.DEGREES:
            case symbols_1.FnSymbol.EXP:
            case symbols_1.FnSymbol.FLOOR:
            case symbols_1.FnSymbol.LOG10:
            case symbols_1.FnSymbol.RADIANS:
            case symbols_1.FnSymbol.ROUND:
            case symbols_1.FnSymbol.SIN:
            case symbols_1.FnSymbol.SQRT:
            case symbols_1.FnSymbol.SQUARE:
            case symbols_1.FnSymbol.SIGN:
            case symbols_1.FnSymbol.TAN:
            case symbols_1.FnSymbol.TRUNC:
            case symbols_1.FnSymbol.IS_ARRAY:
            case symbols_1.FnSymbol.IS_BOOL:
            case symbols_1.FnSymbol.IS_DEFINED:
            case symbols_1.FnSymbol.IS_NULL:
            case symbols_1.FnSymbol.IS_NUMBER:
            case symbols_1.FnSymbol.IS_OBJECT:
            case symbols_1.FnSymbol.IS_PRIMITIVE:
            case symbols_1.FnSymbol.IS_STRING:
            case symbols_1.FnSymbol.LENGTH:
            case symbols_1.FnSymbol.LOWER:
            case symbols_1.FnSymbol.LTRIM:
            case symbols_1.FnSymbol.REVERSE:
            case symbols_1.FnSymbol.RTRIM:
            case symbols_1.FnSymbol.TRIM:
            case symbols_1.FnSymbol.UPPER:
            case symbols_1.FnSymbol.TO_STRING:
            case symbols_1.FnSymbol.ARRAY_LENGTH:
            case symbols_1.FnSymbol.ST_ISVALID:
            case symbols_1.FnSymbol.ST_ISVALIDDETAILED:
                return serializers.unaryFn(expression, paramIndex, isProjection);
            case symbols_1.FnSymbol.ATN2:
            case symbols_1.FnSymbol.LOG:
            case symbols_1.FnSymbol.POWER:
            case symbols_1.FnSymbol.CONTAINS:
            case symbols_1.FnSymbol.ENDSWITH:
            case symbols_1.FnSymbol.STARTSWITH:
            case symbols_1.FnSymbol.INDEX_OF:
            case symbols_1.FnSymbol.LEFT:
            case symbols_1.FnSymbol.REPLICATE:
            case symbols_1.FnSymbol.RIGHT:
            case symbols_1.FnSymbol.ST_DISTANCE:
            case symbols_1.FnSymbol.ST_WITHIN:
            case symbols_1.FnSymbol.ST_INTERSECTS:
                return serializers.binaryFn(expression, paramIndex, isProjection);
            case symbols_1.FnSymbol.REPLACE:
            case symbols_1.FnSymbol.SUBSTRING:
            case symbols_1.FnSymbol.ARRAY_CONTAINS:
            case symbols_1.FnSymbol.ARRAY_SLICE:
                return serializers.ternaryFn(expression, paramIndex, isProjection);
            case symbols_1.FnSymbol.ARRAY_CONCAT:
                return serializers.arrayConcat(expression, paramIndex, isProjection);
            default:
                throw new TypeError('Unrecognized Cosmos DB SQL query builder');
        }
    }
    {
        // This means we have a value rather than a SQL expression. Simply throw
        // the value in the next available param
        const param = `@p${paramIndex}`;
        return {
            query: param,
            parameters: [
                {
                    name: param,
                    // TODO: The Cosmos DB docs and runtime behavior suggest that
                    // all JSON types are allowed, but the library only lists
                    // primitive types in its Typescript declarations:
                    //
                    //   https://github.com/Azure/azure-cosmos-js/issues/233
                    //
                    value: expression
                }
            ]
        };
    }
}
exports.default = serializeExpression;
var serializers;
(function (serializers) {
    function property(expr, paramIndex, isProjection) {
        const [, first, rest] = expr;
        const propsStr = [first, ...rest]
            .map((prop, index) => {
            if (index === 0 && typeof prop !== 'string') {
                throw new TypeError('Cosmos DB property paths must begin with a string');
            }
            if (typeof prop !== 'string' && typeof prop !== 'number') {
                throw new TypeError('Cosmos DB property names must be strings or numbers');
            }
            // We just JSON.stringify() the property once to get the string
            // form for numbers or the quoted string for strings
            return `[${JSON.stringify(prop)}]`;
        })
            .join('');
        // _ts is a special snowflake that we have to deal with because its time is
        // represented in seconds, not milliseconds. As a result, if we find that
        // the property is _ts, we multiply it by 1000 for use in expressions, but
        // not projections.
        if (!isProjection && propsStr === '["_ts"]') {
            return {
                query: `(c${propsStr} * 1000)`,
                parameters: []
            };
        }
        return {
            query: `c${propsStr}`,
            parameters: []
        };
    }
    serializers.property = property;
    function binaryLogical(expr, paramIndex, isProjection) {
        const [sym, expressions] = expr;
        return op(sym, expressions, paramIndex, isProjection);
    }
    serializers.binaryLogical = binaryLogical;
    function unaryLogical(expr, paramIndex, isProjection) {
        const [, expression] = expr;
        let runningParamIndex = paramIndex;
        const { query, parameters } = serializeExpression(expression, runningParamIndex, isProjection);
        runningParamIndex += parameters.length;
        return {
            query: `(${symbols_1.OpSymbol.getOp(expr[0])} ${query})`,
            parameters
        };
    }
    serializers.unaryLogical = unaryLogical;
    function binaryOperator(expr, paramIndex, isProjection) {
        const [sym, left, right] = expr;
        return op(sym, [left, right], paramIndex, isProjection);
    }
    serializers.binaryOperator = binaryOperator;
    function ternaryOperator(expr, paramIndex, isProjection) {
        const [, test, first, second] = expr;
        let runningParamIndex = paramIndex;
        const { query: testQuery, parameters: testParams } = serializeExpression(test, runningParamIndex, isProjection);
        runningParamIndex += testParams.length;
        const { query: firstQuery, parameters: firstParams } = serializeExpression(first, runningParamIndex, isProjection);
        runningParamIndex += firstParams.length;
        const { query: secondQuery, parameters: secondParams } = serializeExpression(second, runningParamIndex, isProjection);
        runningParamIndex += secondParams.length;
        return {
            query: `(${testQuery} ? ${firstQuery} : ${secondQuery})`,
            parameters: [...testParams, ...firstParams, ...secondParams]
        };
    }
    serializers.ternaryOperator = ternaryOperator;
    function inOp(expr, paramIndex, isProjection) {
        const [, value, candidates] = expr;
        let runningParamIndex = paramIndex;
        const { query: valueQuery, parameters: valueParams } = serializeExpression(value, runningParamIndex, isProjection);
        runningParamIndex += valueParams.length;
        const { subqueries, parameters } = candidates
            .filter(expr => expr !== undefined)
            .map(expr => {
            const result = serializeExpression(expr, runningParamIndex, isProjection);
            runningParamIndex += result.parameters.length;
            return result;
        })
            .reduce(mergeSubqueries, { subqueries: [], parameters: [] });
        return {
            query: `(${valueQuery} ${symbols_1.OpSymbol.getOp(expr[0])} (${subqueries.join(', ')}))`,
            parameters: [...valueParams, ...parameters]
        };
    }
    serializers.inOp = inOp;
    function concat(expr, paramIndex, isProjection) {
        const [, first, second, rest] = expr;
        return op(expr[0], [first, second, ...rest], paramIndex, isProjection);
    }
    serializers.concat = concat;
    function pi(expr, paramIndex, isProjection) {
        return fn(expr[0], [], paramIndex, isProjection);
    }
    serializers.pi = pi;
    function unaryFn(expr, paramIndex, isProjection) {
        const [sym, expression] = expr;
        return fn(sym, [expression], paramIndex, isProjection);
    }
    serializers.unaryFn = unaryFn;
    function binaryFn(expression, paramIndex, isProjection) {
        const [sym, first, second] = expression;
        return fn(sym, [first, second], paramIndex, isProjection);
    }
    serializers.binaryFn = binaryFn;
    function ternaryFn(expression, paramIndex, isProjection) {
        const [sym, first, second, third] = expression;
        return fn(sym, [first, second, third], paramIndex, isProjection);
    }
    serializers.ternaryFn = ternaryFn;
    function arrayConcat(expression, paramIndex, isProjection) {
        const [, first, second, rest] = expression;
        return fn(expression[0], [first, second, ...rest], paramIndex, isProjection);
    }
    serializers.arrayConcat = arrayConcat;
    function op(sym, expressions, paramIndex, isProjection) {
        let runningParamIndex = paramIndex;
        const { subqueries, parameters } = expressions
            .filter(expr => expr !== undefined)
            .map(expr => {
            const result = serializeExpression(expr, runningParamIndex, isProjection);
            runningParamIndex += result.parameters.length;
            return result;
        })
            .reduce(mergeSubqueries, { subqueries: [], parameters: [] });
        return {
            query: `(${subqueries.join(` ${symbols_1.OpSymbol.getOp(sym)} `)})`,
            parameters
        };
    }
    function fn(sym, expressions, paramIndex, isProjection) {
        let runningParamIndex = paramIndex;
        const { subqueries, parameters } = expressions
            .filter(expr => expr !== undefined)
            .map(expr => {
            const result = serializeExpression(expr, runningParamIndex);
            runningParamIndex += result.parameters.length;
            return result;
        })
            .reduce(mergeSubqueries, { subqueries: [], parameters: [] });
        return {
            query: `${symbols_1.FnSymbol.getName(sym)}(${subqueries.join(', ')})`,
            parameters
        };
    }
})(serializers || (serializers = {}));
function isCustom(expression) {
    return (Array.isArray(expression) &&
        expression.length >= 1 &&
        typeof expression[0] === 'symbol');
}
function mergeSubqueries(acc, val) {
    return {
        subqueries: [...acc.subqueries, val.query],
        parameters: [...acc.parameters, ...val.parameters]
    };
}
//# sourceMappingURL=expression.js.map