"use strict";
/*!
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
function lock(lockDuration, maxCount, lockToken) {
    const lockDurationMs = Number(lockDuration);
    const maxCountNum = Number(maxCount);
    const context = getContext();
    const container = context.getCollection();
    const self = container.getSelfLink();
    const response = context.getResponse();
    const now = Date.now();
    container.queryDocuments(self, {
        query: `SELECT TOP @count *
            FROM c
            WHERE c.config.enabled = true AND
                  c.config.nextRunTime < @now AND
                  c.config.lockedUntilTime < @now
            ORDER BY c.config.nextRunTime ASC`,
        parameters: [
            {
                name: '@count',
                value: maxCountNum
            },
            {
                name: '@now',
                value: now
            }
        ]
    }, {}, (err, tasks) => {
        if (err) {
            return context.abort(err);
        }
        tryLockTasks(tasks);
    });
    function tryLockTasks(tasks) {
        if (tasks.length === 0) {
            return response.setBody({ tasks, hasMore: false });
        }
        let completedCount = 0;
        const results = [];
        for (let i = 0; i < tasks.length; i += 1) {
            results.push(undefined);
            // We have to pass through the index and pull it back out of the cb
            // because the number is not passed to the closure correctly in
            // Cosmos DB sprocs.
            tryLockTask(tasks, i, (err, task, index) => {
                // Populate the task if we were able to lock it
                if (!err) {
                    results[index] = task;
                }
                completedCount += 1;
                if (completedCount >= tasks.length) {
                    const lockedTasks = results.filter(result => result);
                    // There may be more available if we couldn't lock some that
                    // we found or we returned everything they asked for
                    const hasMore = lockedTasks.length < tasks.length || tasks.length === maxCountNum;
                    return response.setBody({ tasks: lockedTasks, hasMore });
                }
            });
        }
    }
    function tryLockTask(tasks, index, cb) {
        const task = tasks[index];
        const now = Date.now();
        // If the lock expired without us being able to release the task from
        // processing explicitly, we need to increment the attempt/delivery
        // counters for that delivery as if it had been finished.
        if (task.config.lockToken) {
            task.config.attempts += 1;
            task.config.deliveries += 1;
        }
        task.config.lockedUntilTime = now + lockDurationMs;
        task.config.lockToken = lockToken;
        // If the task does not already have a current start time set, set one now
        if (!task.config.currentRunStartTime) {
            task.config.currentRunStartTime = now;
        }
        const accepted = container.replaceDocument(task._self, task, { etag: task._etag }, (err, updatedTask) => cb(err, Object.assign({}, task, updatedTask), index));
        if (!accepted) {
            cb(new Error('not accepted'), undefined, index);
        }
    }
}
exports.lock = lock;
const def = {
    id: 'lock-v1',
    fn: lock
};
exports.default = def;
//# sourceMappingURL=lock.js.map