"use strict";
/*!
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const util = __importStar(require("util"));
const retry_1 = require("retry");
const uuid_1 = require("uuid");
const constants_1 = require("../constants");
const error_1 = __importStar(require("../error"));
const interceptor_1 = __importDefault(require("../interceptor"));
const computeNextRun_1 = __importDefault(require("../utils/computeNextRun"));
const data_1 = __importDefault(require("./data"));
const types_1 = require("./types");
class ActiveTaskImpl extends events_1.EventEmitter {
    constructor(data, interceptor, options) {
        super();
        /**
         * Internal processing state. DO NOT USE DIRECTLY.
         */
        this._internalProcessingState = types_1.ProcessingState.Active;
        this._finishPatches = {
            completeRun: (succeeded) => ({
                runs: this.runs + 1,
                attempts: 0,
                deliveries: 0,
                lastRun: {
                    startTime: this.currentRunStartTime
                        ? this.currentRunStartTime.getTime()
                        : this._processingStart,
                    finishTime: this._data.timestamp(),
                    succeeded
                },
                currentRunStartTime: undefined
            }),
            scheduleNextRun: (nextRunTime) => ({
                nextRunTime
            }),
            consumeAttempt: () => ({
                attempts: this.attempts + 1
            }),
            consumeDelivery: () => ({
                deliveries: this.deliveries + 1
            })
        };
        /**
         * Attempt to renew the lock on the task
         */
        this._renewLock = async () => {
            const lockToken = uuid_1.v4();
            this._nextLockToken = lockToken;
            // Continually attempt to renew the lock as long as we are still able to
            // renew the lock.
            while (this._canRenewLock) {
                let cancelled = false;
                try {
                    await this._interceptor.task(this, interceptor_1.default.TaskOperation.RenewLock, this._data.ref, async () => {
                        const cancel = Symbol('CANCEL');
                        try {
                            return await this._data.patch(() => {
                                if (!this._canRenewLock) {
                                    throw cancel;
                                }
                                return {
                                    lockedUntilTime: this._data.timestamp() + this._options.lockDurationMs,
                                    lockToken
                                };
                            }, true);
                        }
                        catch (err) {
                            // We ignore the error, but mark it cancelled if it is our cancel
                            // symbol. If not, we re-throw the error.
                            if (err === cancel) {
                                cancelled = true;
                                return {
                                    result: undefined,
                                    ruConsumption: undefined
                                };
                            }
                            throw err;
                        }
                    });
                    // Once we have renewed the lock, store the new lock token so we can be
                    // sure we own the lock, schedule the next renewal, and clear out of the
                    // renewal loop.
                    if (!cancelled) {
                        this._lockToken = lockToken;
                        this.emit('lockRenewed');
                        this._scheduleRenewLock();
                        return;
                    }
                }
                catch (err) {
                    // Swallow errors
                }
                finally {
                    // Clear out the next lock token tracker if appropriate
                    // regardless of the outcome
                    if (this._nextLockToken === lockToken) {
                        this._nextLockToken = undefined;
                    }
                    // We break out of our loop for cancelled operations
                    if (cancelled) {
                        return;
                    }
                }
            }
        };
        /**
         * Poll the database for updates to this task. This is used to keep etag
         * versions in sync and handle task disabling.
         */
        this._poll = async () => {
            await this._interceptor.task(this, interceptor_1.default.TaskOperation.Refresh, this._data.ref, async () => this._data.sync());
        };
        /**
         * Signal that the lock was lost, or schedule a future check if necessary
         */
        this._checkLock = () => {
            if (!this.locked &&
                (this.processingState === types_1.ProcessingState.Active ||
                    this.processingState === types_1.ProcessingState.Finishing)) {
                // If we're not locked any more, signal that we lost the lock
                this._processingState = types_1.ProcessingState.LockLost;
                this._markFinished(types_1.ProcessingResult.LockLost);
                this.emit('lockLost');
                this._cleanup();
            }
            else {
                // If we still have a lock, schedule when we need to check that the
                // lock has expired
                this._scheduleLockExpiration();
            }
        };
        /**
         * Event handler for verifying task state whenever a server update happens
         */
        this._onDataUpdated = () => {
            // We bump the checks onto the next loop so that any processes that
            // follow directly from the update can go first.
            setTimeout(() => {
                // Someone else may have locked our task
                this._checkLock();
                // A user mayu have disabled this task
                this._checkDisabled();
            });
        };
        // We save some data off to the side so we know when it was changed by
        // someone else.
        this._lockToken = data.document.config.lockToken;
        this._enabled = data.document.config.enabled;
        this._data = data;
        this._interceptor = interceptor;
        this._options = options;
        this._processingStart = this._data.timestamp();
        this._schedulePoll();
        this._scheduleRenewLock();
        this._scheduleLockExpiration();
        this._data.on('updated', this._onDataUpdated);
    }
    static create(client, interceptor, document, options) {
        return new ActiveTaskImpl(new data_1.default(client, document), interceptor, options);
    }
    get id() {
        return this._data.id;
    }
    get type() {
        return this._data.type;
    }
    get status() {
        return this._data.status;
    }
    get enabled() {
        return this._data.enabled;
    }
    get createTime() {
        return this._data.createTime;
    }
    get nextRunTime() {
        return this._data.nextRunTime;
    }
    get lastUpdatedTime() {
        return this._data.lastUpdatedTime;
    }
    get lastRun() {
        return this._data.lastRun;
    }
    get currentRunStartTime() {
        return this._data.currentRunStartTime;
    }
    get deliveries() {
        return this._data.deliveries;
    }
    get attempts() {
        return this._data.attempts;
    }
    get runs() {
        return this._data.runs;
    }
    get payload() {
        return this._data.payload;
    }
    set payload(payload) {
        this._data.payload = payload;
    }
    get interval() {
        return this._data.interval;
    }
    get lastRunTime() {
        return this._data.lastRunTime;
    }
    get locked() {
        return ((this._data.document.config.lockToken === this._lockToken ||
            this._data.document.config.lockToken === this._nextLockToken) &&
            this._data.timestamp() < this._data.document.config.lockedUntilTime);
    }
    get processingState() {
        return this._internalProcessingState;
    }
    /**
     * The maximum amount of time in milliseconds to run the task.
     */
    get _maxExecutionTimeMs() {
        return this._data.document.config.maxExecutionTimeMs !== undefined
            ? this._data.document.config.maxExecutionTimeMs
            : this._options.maxExecutionTimeMs;
    }
    /**
     * Indicates whether the lock can/should be renewed.
     */
    get _canRenewLock() {
        return (this.locked &&
            (this._maxExecutionTimeMs === 0 ||
                this._data.document.config.lockedUntilTime <
                    this._processingStart + this._maxExecutionTimeMs) &&
            this.processingState !== types_1.ProcessingState.Finished &&
            this.processingState !== types_1.ProcessingState.LockLost);
    }
    /**
     * Setter for updating the processing state, which also notifies clients
     * of any changes.
     */
    set _processingState(state) {
        if (state !== this._internalProcessingState) {
            this._internalProcessingState = state;
            this.emit('stateChanged', this._internalProcessingState);
        }
    }
    async save() {
        await this._interceptor.task(this, interceptor_1.default.TaskOperation.Save, this._data.ref, async () => this._data.patchWithPayload(() => {
            this._assertProcessing();
            return {};
        }));
    }
    async delete() {
        await this._interceptor.task(this, interceptor_1.default.TaskOperation.Delete, this._data.ref, async () => this._finish(types_1.ProcessingResult.Delete, async () => this._data.delete(() => this._assertProcessing())));
    }
    async complete(nextRunDelayMs, savePayload) {
        await this._interceptor.task(this, interceptor_1.default.TaskOperation.Complete, this._data.ref, async () => this._acknowledge(types_1.ProcessingResult.Complete, Object.assign(Object.assign({}, this._finishPatches.completeRun(true)), this._finishPatches.scheduleNextRun(nextRunDelayMs !== undefined
            ? this._data.timestamp() + nextRunDelayMs
            : computeNextRun_1.default(this.interval, this._processingStart, this.lastRunTime))), savePayload, { delayMs: nextRunDelayMs }));
    }
    async fail(err, nextRunDelayMs, savePayload) {
        await this._interceptor.task(this, interceptor_1.default.TaskOperation.Fail, this._data.ref, async () => this._acknowledge(types_1.ProcessingResult.Fail, Object.assign(Object.assign({}, this._finishPatches.completeRun(false)), this._finishPatches.scheduleNextRun(nextRunDelayMs !== undefined
            ? nextRunDelayMs + this._data.timestamp()
            : computeNextRun_1.default(this.interval, this._processingStart, this.lastRunTime))), savePayload, { delayMs: nextRunDelayMs, error: err }));
    }
    async retry(err, delayMs, savePayload) {
        // If we've exceeded the maximum number of attempts, fail.
        if (this.attempts >= this._options.retries.retries) {
            return await this.fail(err, undefined, savePayload);
        }
        await this._interceptor.task(this, interceptor_1.default.TaskOperation.Retry, this._data.ref, async () => {
            const resolvedDelayMs = delayMs !== undefined
                ? delayMs
                : retry_1.createTimeout(this.attempts, this._options.retries);
            return await this._acknowledge(types_1.ProcessingResult.Retry, Object.assign(Object.assign(Object.assign({}, this._finishPatches.scheduleNextRun(this._data.timestamp() + resolvedDelayMs)), this._finishPatches.consumeDelivery()), this._finishPatches.consumeAttempt()), savePayload, { delayMs: resolvedDelayMs, error: err });
        });
    }
    async defer(delayMs, savePayload) {
        await this._interceptor.task(this, interceptor_1.default.TaskOperation.Defer, this._data.ref, async () => this._acknowledge(types_1.ProcessingResult.Defer, Object.assign(Object.assign({}, this._finishPatches.scheduleNextRun(this._data.timestamp() + delayMs)), this._finishPatches.consumeDelivery()), savePayload, { delayMs }));
    }
    async release(savePayload) {
        await this._interceptor.task(this, interceptor_1.default.TaskOperation.Release, this._data.ref, async () => this._acknowledge(types_1.ProcessingResult.Release, Object.assign({}, this._finishPatches.consumeDelivery()), savePayload));
    }
    forceRelease(err) {
        this._markFinished(types_1.ProcessingResult.ForceRelease, undefined, err);
        this._cleanup();
    }
    /**
     * Acknowledge the task and finish processing using the given changes.
     *
     * @param result      - The processing result if the acknowledge succeeds
     * @param changes     - The patch to make to the task configuration
     * @param savePayload - Whether or not to save the payload (default: true)
     * @param logMeta     - Optional metadata for logging purposes
     */
    async _acknowledge(result, changes, savePayload, logMeta = {}) {
        // Add base changes to make for all types of finishes
        const resolvedChanges = Object.assign(Object.assign({}, changes), { lockedUntilTime: 0, lockToken: undefined });
        return this._finish(result, async () => savePayload || savePayload === undefined
            ? this._data.patchWithPayload(() => {
                this._assertProcessing();
                return resolvedChanges;
            })
            : this._data.patch(() => {
                this._assertProcessing();
                return resolvedChanges;
            }), logMeta);
    }
    /**
     * Run the given operation to finish processing (and log the processing).
     *
     * @param result      - The processing result if the finish succeeds
     * @param finisher    - The operation to run when actually finishing the
     *                      task
     * @param logMeta     - Optional metadata for logging purposes
     */
    async _finish(result, finisher, logMeta = {}) {
        this._assertProcessing();
        this._assertNotFinishing();
        this._processingState = types_1.ProcessingState.Finishing;
        try {
            const response = await finisher();
            // If the call succeeded, we're finished
            if (this.processingState === types_1.ProcessingState.Finishing) {
                this._processingState = types_1.ProcessingState.Finished;
            }
            this._markFinished(result, logMeta.delayMs, logMeta.error);
            this._cleanup();
            return response;
        }
        catch (e) {
            // If we're still finishing and the call failed, set it back to
            // active
            if (this.processingState === types_1.ProcessingState.Finishing) {
                this._processingState = types_1.ProcessingState.Active;
            }
            throw e;
        }
    }
    /**
     * Notify consumers that the task is finished processing.
     *
     * @param result  - The processing result
     * @param delayMs - Optional delay before reprocessing
     * @param error   - Optional error from processing
     */
    _markFinished(result, delayMs, error) {
        this.emit('finished', { result, error, delayMs });
    }
    /**
     * Schedule the next poll attempt.
     */
    _schedulePoll() {
        // Make sure we don't start stacking timers
        if (this._pollTimer) {
            clearTimeout(this._pollTimer);
            this._pollTimer = undefined;
        }
        // We only poll if the task is still enabled, since the only thing we're
        // looking for is whether the task is disabled. We also must have a
        // positive refresh interval (an interval of 0 disables the feature)
        if (this.enabled && this._options.refreshIntervalMs > 0) {
            this._pollTimer = setTimeout(this._poll, this._options.refreshIntervalMs);
        }
    }
    /**
     * Schedule the next lock renewal.
     */
    _scheduleRenewLock() {
        // Make sure we don't start stacking timers
        if (this._lockTimer) {
            clearTimeout(this._lockTimer);
            this._lockTimer = undefined;
        }
        // We only schedule the lock renewal if we can actually still renew the
        // lock for this task.
        if (this._canRenewLock) {
            // Amount of time before lock expiration to renew
            const gracePeriod = this._options.lockDurationMs * (1 - constants_1.LOCK_RATIO);
            // Time when we should start renewing the lock
            const deadline = this._data.document.config.lockedUntilTime - gracePeriod;
            // Amount of time to wait, computed as time when we want to start
            // minus the current time.
            const delayMs = Math.max(deadline - this._data.timestamp(), 0);
            this._lockTimer = setTimeout(this._renewLock, delayMs);
        }
    }
    /**
     * Set up the timer that indicates that the lock is expired.
     */
    _scheduleLockExpiration() {
        // Make sure we don't start stacking timers
        if (this._lockExpirationTimer) {
            clearTimeout(this._lockExpirationTimer);
            this._lockExpirationTimer = undefined;
        }
        if (this.processingState === types_1.ProcessingState.Active ||
            this.processingState === types_1.ProcessingState.Finishing) {
            this._lockExpirationTimer = setTimeout(this._checkLock, Math.max(this._data.document.config.lockedUntilTime - this._data.timestamp(), 0));
        }
    }
    /**
     * Check whether the task has been disabled
     */
    _checkDisabled() {
        if (this._enabled && !this.enabled) {
            this.emit('disabled');
            this._enabled = false;
        }
    }
    /**
     * Clean up any timers, listeners, etc. that are active during task
     * processing.
     */
    _cleanup() {
        if (this.processingState === types_1.ProcessingState.Active ||
            this.processingState === types_1.ProcessingState.Finishing) {
            this._processingState = types_1.ProcessingState.Finished;
        }
        this.emit('inactive');
        // Clear out any timers that are running
        if (this._pollTimer) {
            clearTimeout(this._pollTimer);
            this._pollTimer = undefined;
        }
        if (this._lockTimer) {
            clearTimeout(this._lockTimer);
            this._lockTimer = undefined;
        }
        if (this._lockExpirationTimer) {
            clearTimeout(this._lockExpirationTimer);
            this._lockExpirationTimer = undefined;
        }
        // Dispose of our underlying data
        this._data.dispose();
        // Remove anything listening to our events as nothing else will be
        // firing after this point
        this.removeAllListeners();
    }
    /**
     * Verify that the task is not finished processing
     */
    _assertProcessing() {
        switch (this.processingState) {
            case types_1.ProcessingState.Finished:
                throw new error_1.default(error_1.ErrorCode.PROCESSING_ALREADY_FINISHED);
            case types_1.ProcessingState.LockLost:
                throw new error_1.default(error_1.ErrorCode.PROCESSING_LOCK_LOST);
        }
    }
    /**
     * Verify that the task is not currently finishing
     */
    _assertNotFinishing() {
        if (this.processingState === types_1.ProcessingState.Finishing) {
            throw new error_1.default(error_1.ErrorCode.PROCESSING_FINISH_IN_PROGRESS);
        }
    }
    toJSON() {
        return Object.assign(Object.assign({}, this._data.toJSON()), { locked: this.locked });
    }
    // We cast as any so that types behave outside of this library
    [util.inspect.custom](depth, opts) {
        return this.toJSON();
    }
}
exports.default = ActiveTaskImpl;
//# sourceMappingURL=active.js.map