"use strict";
/*!
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const query_1 = require("../query");
const types_1 = require("./types");
/**
 * Expressions representing various properties of a task useful for filtering
 * or projection expressions in operations like {@link TaskClient.list}, {@link
 * TaskClient.listSummary}, {@link TaskClient.iterate}, {@link
 * TaskClient.iterateSummary} and more.
 *
 * @public
 */
var t;
(function (t) {
    /**
     * Unique identifier for the task
     *
     * @public
     */
    t.id = query_1.p.str('id');
    /**
     * The type (grouping) for the task
     *
     * @public
     */
    t.type = query_1.p.str('config', 'type');
    /**
     * Indicates whether the task is eligible for processing
     *
     * @public
     */
    t.enabled = query_1.p.bool('config', 'enabled');
    /**
     * Unix ms epoch representing the time when the task was first created
     *
     * @public
     */
    t.createTime = query_1.p.num('config', 'createTime');
    /**
     * Unix ms epoch representing the next time the task should be processed.
     * Can be earlier than the current time if the task is pending or currently
     * running and is undefined if the task is not running and has no future
     * runs scheduled.
     *
     * @public
     */
    t.nextRunTime = query_1.p.num('config', 'nextRunTime');
    /**
     * Unix ms epoch representing the last time the task was updated, either
     * explicitly or as part of processing.
     */
    t.lastUpdatedTime = query_1.p.num('_ts');
    /**
     * Metadata about the most recent completed/failed run of the task. If the
     * task has neither failed a run, this is undefined.
     *
     * @public
     */
    t.lastRun = query_1.p.obj('config', 'lastRun');
    /**
     * Unix ms epoch representing when the most recent run began processing. If
     * the task has neither completed nor failed a run, this is undefined.
     *
     * @public
     */
    t.lastRunStartTime = query_1.p.num('config', 'lastRun', 'startTime');
    /**
     * Unix ms epoch representing when the last run finished processing. If the
     * task has neither completed nor failed a run, this is undefined.
     *
     * @public
     */
    t.lastRunFinishTime = query_1.p.num('config', 'lastRun', 'finishTime');
    /**
     * Boolean indicating whether or not the most recent run succeeded or
     * failed. If the task has neither completed nor failed a run, this is
     * undefined.
     *
     * @public
     */
    t.lastRunSucceeded = query_1.p.bool('config', 'lastRun', 'succeeded');
    /**
     * Number of times the task has been delivered for processing. This number
     * only includes deliveries that are not currently running and resets after
     * each time the task completes or fails a run.
     *
     * @public
     */
    t.deliveries = query_1.p.num('config', 'deliveries');
    /**
     * Number of times task processing has been attempted unsuccessfully. This
     * number only includes attempts that are not currently running and resets
     * after each time the task completes or fails a run. It is similar to
     * deliveries but only increments for results that indicate some sort of
     * failure. These include calling {@link ActiveTask.retry} and when the
     * task's lock expires.
     *
     * @public
     */
    t.attempts = query_1.p.num('config', 'attempts');
    /**
     * Number of times task processing has run to completion, regardless of
     * whether they ultimately succeeded or failed. This number only includes
     * previously completed runs. For one-time tasks, this will only ever be 0
     * or 1, but can be arbitrarily large for recurring tasks that have run many
     * times.
     *
     * @public
     */
    t.runs = query_1.p.num('config', 'runs');
    /**
     * If defined, the schedule on which to run the task. If it is a number, it
     * represents the number of milliseconds between each run of the task. If it
     * is a string, it is a cron string indicating the schedule on which to run.
     *
     * @public
     */
    t.interval = query_1.p.base('config', 'interval');
    /**
     * User-defined payload holding information about the task. If called with
     * no arguments, it points to the entire payload. However, you can also
     * reference portions of the payload by providing arguments to the function.
     * Each argument represents a property name or array index that, when
     * combined, form a path to a property in the payload.
     *
     * @example Retrieve the full payload
     *
     * ```ts
     * t.payload()
     * ```
     *
     * @example Retrieve payload.prop[2]
     *
     * ```ts
     * t.payload('prop', 2)
     * ```
     *
     * @public
     */
    t.payload = (...path) => query_1.p.base('payload', ...path);
    const lockedUntilTime = query_1.p.num('config', 'lockedUntilTime');
    /**
     * Status is a derived property. It does not behave as a property because it
     * is computed dynamically, but this helper can be used to determine whether
     * the status belongs to each of the possible states.
     *
     * @param status  - Status to check the task for
     *
     * @public
     */
    t.hasStatus = (status) => {
        const now = Date.now();
        switch (status) {
            case types_1.TaskStatus.Scheduled:
                // Enabled AND we haven't reached the nextRunTime
                return query_1.q.and(t.enabled, query_1.q.greaterThan(t.nextRunTime, now));
            case types_1.TaskStatus.Pending:
                return query_1.q.and(
                // The job is enabled
                t.enabled, 
                // We have passed the next run time
                query_1.q.lessThanOrEqual(t.nextRunTime, now), 
                // But there is no active lock
                query_1.q.lessThan(lockedUntilTime, now));
            case types_1.TaskStatus.Running:
                return query_1.q.and(
                // The job is enabled
                t.enabled, 
                // We have passed the next run time
                query_1.q.lessThanOrEqual(t.nextRunTime, now), 
                // And there is an active lock
                query_1.q.greaterThanOrEqual(lockedUntilTime, now));
            case types_1.TaskStatus.Completed:
                return query_1.q.and(
                // We don't have a next run
                query_1.q.not(query_1.q.isDefined(t.nextRunTime)), 
                // And the last run succeeded (also implies there was a last run)
                t.lastRunSucceeded);
            case types_1.TaskStatus.Failed:
                return query_1.q.and(
                // We don't have a next run
                query_1.q.not(query_1.q.isDefined(t.nextRunTime)), 
                // And either we don't have a last run, or the last run didn't succeed
                query_1.q.or(query_1.q.not(query_1.q.isDefined(t.lastRun)), query_1.q.not(t.lastRunSucceeded)));
            case types_1.TaskStatus.Disabling:
                return query_1.q.and(
                // The job is disabled
                query_1.q.not(t.enabled), 
                // There is a next run
                query_1.q.isDefined(t.nextRunTime), 
                // The next run has passed
                query_1.q.lessThanOrEqual(t.nextRunTime, now), 
                // And there is an active lock
                query_1.q.greaterThanOrEqual(lockedUntilTime, now));
            case types_1.TaskStatus.Disabled:
                return query_1.q.and(
                // The job is disabled
                query_1.q.not(t.enabled), 
                // There is a next run
                query_1.q.isDefined(t.nextRunTime), query_1.q.or(
                // Either we haven't reached the next run time
                query_1.q.greaterThan(t.nextRunTime, now), 
                // Or there is no active lock
                query_1.q.lessThan(lockedUntilTime, now)));
        }
    };
})(t || (t = {}));
exports.default = t;
//# sourceMappingURL=properties.js.map