"use strict";
/*!
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("./types");
/**
 * Computes the current status of a task from the task document information.
 *
 * @param config    The configuration portion of the task document
 * @param now       The current unix ms time to use for computing task state
 */
function taskStatus(config, now) {
    const status = enabledTaskStatus(config, now);
    // Terminal states (Completed, Failed) override disabled
    if (config.enabled ||
        status === types_1.TaskStatus.Completed ||
        status === types_1.TaskStatus.Failed) {
        return status;
    }
    // If the task is running and it's disabled, it's disabling
    if (status === types_1.TaskStatus.Running) {
        return types_1.TaskStatus.Disabling;
    }
    // If the task is disabled and not running, it's fully disabled
    return types_1.TaskStatus.Disabled;
}
exports.default = taskStatus;
/**
 * Returns true if the task is in a terminal state.
 *
 * @param config    The configuration portion of the task document
 * @param now       The current unix ms time to use for computing task state
 */
function isFinished(config, now) {
    const status = taskStatus(config, now);
    return status === types_1.TaskStatus.Completed || status === types_1.TaskStatus.Failed;
}
exports.isFinished = isFinished;
/**
 * Computes the current state of a task from the task document, excluding
 * disabled states.
 *
 * @param config    The configuration portion of the task document
 */
function enabledTaskStatus(config, now) {
    if (config.nextRunTime === undefined) {
        if (config.lastRun && config.lastRun.succeeded) {
            // No next run and a successful last run means we completed
            return types_1.TaskStatus.Completed;
        }
        // No next run and a failed last run means we failed
        return types_1.TaskStatus.Failed;
    }
    if (config.lockedUntilTime > now) {
        // We have a valid lock on the task. Set it to running
        return types_1.TaskStatus.Running;
    }
    if (config.nextRunTime < now) {
        // If we have reached the next run time and we don't have a lock, task
        // processing is pending
        return types_1.TaskStatus.Pending;
    }
    // If we haven't yet reached the next run time and we don't have a lock,
    // the task is scheduled
    return types_1.TaskStatus.Scheduled;
}
//# sourceMappingURL=status.js.map